#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <jmp.h>
#include <jthread.h>
#include <cls.h>
#include <method.h>

/** Thread comparators */

/** Compare two jthreads by there name. */
int jthread_compr_name (const void* v1, const void* v2) {
    jthread **t1 = (jthread**) v1;
    jthread **t2 = (jthread**) v2;
    return strcmp (jthread_get_thread_name(*t1), jthread_get_thread_name(*t2));
}

/** compare threads based on thread group */
int jthread_compr_group (const void* v1, const void* v2) {
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    return (strcmp (jthread_get_group_name (*t1), jthread_get_group_name (*t2)));
}

/** compare threads based on thread parent */
int jthread_compr_parent (const void* v1, const void* v2) {
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    return (strcmp (jthread_get_parent_name (*t1), jthread_get_parent_name (*t2)));
}

/** Compare threads based on contenation time for the threads. */
int jthread_compr_contenation (const void* v1, const void* v2) {
    long l;
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    if ((l = ((*t1)->contended.tv_sec - (*t2)->contended.tv_sec)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*t2)->contended.tv_usec - (*t1)->contended.tv_usec;
    }
}


/** class comparators... */

/** compare classes based on class name */
int cls_compr_name (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (strcmp (cls_get_name (*c1), cls_get_name (*c2)));
}

/** compare classes based on instance count */
int cls_compr_instance (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_instances (*c2) - cls_get_instances (*c1));
}

/** compare classes based on maximum instance count */
int cls_compr_max_instance (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_max_instances (*c2) - cls_get_max_instances (*c1));
}

/** compare classes based on current size in bytes */
int cls_compr_size (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_size (*c2) - cls_get_size (*c1));
}


/** Method comparators. */

/** Compare methods based on signature. */
int method_compr_signature (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    m1 = (method**)v1;
    m2 = (method**)v2;
    return (strcmp (method_get_method_signature (*m1), method_get_method_signature (*m2)));
}

/** Compare methods based on method name (and signature if needed) */
int method_compr_name (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    int c = 0; 
    m1 = (method**)v1;
    m2 = (method**)v2;
    c = (strcmp (method_get_method_name (*m1), method_get_method_name (*m2)));
    if (c) 
	return c;
    return method_compr_signature (v1, v2);
}

/** Compare method based on class name, method name and signature (in that order) */
int method_compr_class (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    cls* c1;
    cls* c2;
    int c;
    m1 = (method**)v1;
    m2 = (method**)v2;
    c1 = method_get_owner (*m1);
    c2 = method_get_owner (*m2);
    c = (strcmp (cls_get_class_name (c1), cls_get_class_name (c2)));    
    if (c)
	return c;
    return method_compr_name (v1, v2);
}

/** Compare methods based on used time in the method. */
int method_compr_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l;
    m1 = (method**)v1;
    m2 = (method**)v2;
    if ((l = ((*m1)->time_used.tv_sec - (*m2)->time_used.tv_sec)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*m2)->time_used.tv_usec - (*m1)->time_used.tv_usec;
    }
}

/** Compare methods based on number of calls. */
int method_compr_calls (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    m1 = (method**)v1;
    m2 = (method**)v2;
    return method_get_calls (*m2) - method_get_calls (*m1);
}

/** Compare methods based on time used in method called from this method. */
int method_compr_hold_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l;
    m1 = (method**)v1;
    m2 = (method**)v2;
    if ((l = ((*m1)->time_used.tv_sec_hold - (*m2)->time_used.tv_sec_hold)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*m2)->time_used.tv_usec_hold - (*m1)->time_used.tv_usec_hold;
    }
}

/** Compare methods based on time used in this method and method called from this method. */
int method_compr_total_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l, lusec1, lsec1, lusec2, lsec2;
    m1 = (method**)v1;
    m2 = (method**)v2;

    lsec1 = (*m1)->time_used.tv_sec + (*m1)->time_used.tv_sec_hold;
    lusec1 = (*m1)->time_used.tv_usec + (*m1)->time_used.tv_usec_hold;
    if (lusec1 > 1000000) {
	lsec1++;
	lusec1 -= 1000000;
    }
    lsec2 = (*m2)->time_used.tv_sec + (*m2)->time_used.tv_sec_hold;
    lusec2 = (*m2)->time_used.tv_usec + (*m2)->time_used.tv_usec_hold;
    if (lusec2 > 1000000) {
	lsec2++;
	lusec2 -= 1000000;
    }    

    if ((l = (lsec1 - lsec2)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return lusec2 - lusec1;
    }
}
