#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <jvmpi.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <jthread.h>
#include <mvector.h>
#include <obj.h>

static char* empty = "<unknown>";

jthread* jthread_new (const char* thread_name, const char* group_name, 
		      const char* parent_name, jobjectID thread_id,
		      JNIEnv* env_id) {
    jthread* c;
    
    c = malloc (sizeof (*c));
    if (c == NULL)
	return NULL;
    
    if (thread_name)
	c->thread_name = strdup (thread_name);
    else 
	c->thread_name = empty;
    if (group_name)
	c->group_name = strdup (group_name);
    else 
	c->group_name = empty;
    if (parent_name)
	c->parent_name = strdup (parent_name);
    else 
	c->parent_name = empty;

    if (c->thread_name == NULL ||
	c->group_name == NULL ||
	c->parent_name == NULL) {
	jthread_free (c);
	return NULL;
    }
    
    c->thread_id = thread_id;
    c->env_id = env_id;
    c->contended.tv_usec = 0;
    c->contended.tv_sec = 0;    
    return c;
}

void jthread_free (jthread* t) {
    if (t == NULL)
	return;
    if (t->parent_name != empty)
	free (t->parent_name);
    if (t->group_name != empty)
	free (t->group_name);
    if (t->thread_name != empty)
	free (t->thread_name);
    free (t);
}

size_t jthread_hash_func (void* c, size_t len) {
    jthread* tp = (jthread*)c;
    return ((long)tp->env_id) % len;
}

int jthread_cmp_func (void* c1, void* c2) {
    jthread* tp1 = (jthread*)c1;
    jthread* tp2 = (jthread*)c2;
    return tp1->env_id != tp2->env_id;
}

JNIEnv* jthread_get_env_id (jthread* t) {
    return t->env_id;
}

void jthread_set_env_id (jthread* t, JNIEnv* env_id) {
    t->env_id = env_id;
}

void jthread_method_entry (timerstack* s, jmethodID method_id, const struct timeval* tv) {
    methodtime* tvc;
    if (s->top == s->max) {
	fprintf (stderr, "time to expand timerstack: (%p, %d, %d)\n", s->times, s->top, s->max);
	timerstack_expand (s);
	fprintf (stderr, "timerstack expanded: (%p, %d, %d)\n", s->times, s->top, s->max);
    }
    
    tvc = s->times + s->top;
    s->top++;
    tvc->tv_sec = tv->tv_sec;
    tvc->tv_sec_hold = 0;
    tvc->tv_usec = tv->tv_usec; 
    tvc->tv_usec_hold = 0;
    tvc->method_id = method_id;
}
static inline void update_method_time (method* m, long sdiff, long usdiff, methodtime* tvc) {
    methodtime* method_total;

    method_total = &(m->time_used);
    method_total->tv_sec += sdiff;
    method_total->tv_usec += usdiff;    
    if (method_total->tv_usec > 1000000) {
	method_total->tv_sec++;
	method_total->tv_usec -= 1000000;
    }
    
    method_total->tv_sec_hold += tvc->tv_sec_hold;
    method_total->tv_usec_hold += tvc->tv_usec_hold;
    if (method_total->tv_usec_hold > 1000000) {
	method_total->tv_sec_hold++;
	method_total->tv_usec_hold -= 1000000;
    }
}

void jthread_method_exit (timerstack* s, method* m, const struct timeval* tv) {
    long sdiff, usdiff; 
    methodtime* tvc = NULL;
    
    if (s->top > 0) {
	s->top--;
	tvc = s->times + s->top;
    }
    if (tvc) {
	/* update this methods total time. */
	sdiff = tv->tv_sec - tvc->tv_sec - tvc->tv_sec_hold;
	usdiff = tv->tv_usec - tvc->tv_usec - tvc->tv_usec_hold;
	while (usdiff < 0) {
	    sdiff--;
	    usdiff += 1000000;
	}
	m->calls++;

	update_method_time (m, sdiff, usdiff, tvc);
	/* notify caller how much time this method took.. */
	sdiff = tv->tv_sec - tvc->tv_sec;
	usdiff = tv->tv_usec - tvc->tv_usec;
	if (usdiff < 0) {
	    sdiff--;
	    usdiff += 1000000;
	}
	
	/* get caller and fill in values. */
	if (s->top > 0)
	    tvc = s->times + s->top - 1;
	else 
	    tvc = NULL;
	if (tvc) {
	    method* mc;	    
	    tvc->tv_sec_hold += sdiff;
	    tvc->tv_usec_hold += usdiff;
	    if (tvc->tv_usec_hold > 1000000) {
		tvc->tv_usec_hold %= 1000000;
		tvc->tv_sec_hold += (tvc->tv_usec_hold / 1000000);
	    }
	    mc = get_method (tvc->method_id);
	    if (mc && mc->called_methods) {
		long i = mvector_search (mc->called_methods, m);
		if (i == -1)
		    mvector_add_method (mc->called_methods, m);
	    }
	} 
	/* if we have no parent method we just ignore for now. 
	   Im not sure if we should take actions */
    } else {
	fprintf (stderr, "jthread_method_exit: stack underflow..\n");
    }
}

/** The given thread has is waiting for a monitor. */
void jthread_contenation_enter (timerstack* s, obj* monitor, const struct timeval* tv, jthread* t) {
    jthread_method_entry (s, NULL, tv);
}

/** The given thread got the monitor it was waiting for. */
void jthread_contenation_entered (timerstack* s, obj* monitor, const struct timeval* tv, jthread* t) {
    methodtime* tvc = NULL;
    
    if (s->top > 0) {
	s->top--;
	tvc = s->times + s->top;
    }
    if (tvc) {
	long sdiff, usdiff; 

	/* no hold values to count with.. */
	sdiff = tv->tv_sec - tvc->tv_sec;       
	usdiff = tv->tv_usec - tvc->tv_usec;
	while (usdiff < 0) {
	    sdiff--;
	    usdiff += 1000000;
	}
	
	t->contended.tv_sec += sdiff;
	t->contended.tv_usec += usdiff;
	if (t->contended.tv_usec >= 1000000) {
	    t->contended.tv_usec -= 1000000;
	    t->contended.tv_sec++;
	}
    } else {
	fprintf (stderr, "jthread_method_exit: stack underflow..\n");
    }    
}

char* jthread_get_thread_name (jthread* t) {
    return (t->thread_name);
}
char* jthread_get_group_name (jthread* t) {
    return (t->group_name);
}
char* jthread_get_parent_name (jthread* t) {
    return (t->parent_name);
}

